# Jeremy Norris
# Rebel.com

package Config::TextKV;

use strict;
use FileHandle;

# Description:
# -----------
# This class provides un-instrusive configuration file manipulation services.
# The configuration file format must be a text file consisting of a series of
# key/value pairs separated by a delimeter, and followed by a LF.

# Terminology:
# -----------
# Delimeter:	the character which separates a key/value pair.
# Disabled:     a valud key/value pair which is commented out. 
# Enabled:	a key/value pair which is not commented out.
# Option:	a valid key/value pair separated by a delimeter.
# Status:	the status of a key/value pair (enabled, disabled).

# Additional Notes:
# ----------------
# The configuration file is opened and parsed when the file is the object is
# constructed.  The configuration file is re-written to disk implicitly when
# the object is destroyed, or explicitly by calling $self->_commit.
# Therefore, any changes made by another user during this time will be lost
# unless a write lock is placed on the configuration file during this time.  

# Future Improvements:
# -------------------
# -Modifiy regexps so that they are commentable. (using /x and breaking them
#  up).


# create new instances of self from either the parent class or another
# instance of self.  The arguments are the configuration file to be
# manipulated, and the delimeter character which the configuration file
# uses.

sub new {

  my $proto = shift;
  my $type = ref($proto) || $proto;
  my $self = {};
  $self->{Filename} = shift;
  $self->{D} = shift; 			# delimeter character
  $self->{C} = shift;			# comment character
  #if($self->{C} eq '#') {		# Since the regexps have the /x
  #  $self->{C} = '\x23';		# modifier, #'s must be prefixed
  #}					# by a \.
  $self->{Status} = {};
  $self->{Line} = {};
  $self->{Value} = {};
  bless $self, $type;
  $self->_parsefile;
  $self;

}

# PRIVATE:  This method is called when a new instance of self is created.
# This method parses the configuration file and exposes all the valid
# key/value pairs via internal variables.

sub _parsefile {

  my $self = shift;

  my($line);
 
  $self->{FH} = FileHandle->new($self->{Filename}, 'r+') 
    || die("Unable to open '$self->{Filename}' for editing.\n"); 

  @{$self->{FileData}} = $self->{FH}->getlines;

  $self->{LineCount} = 0;
  foreach $line (@{$self->{FileData}}) {
    $_ = $line; 
    if(s{^\s*$self->{C}\s*(\S+)\s*$self->{D}\s*(\S+)\s*$}
        {$self->{C}$1$self->{D}$2\n}) {
      $self->{Status}->{$1} = 0; # disabled
      $self->{Line}->{$1} = $self->{LineCount};
      $self->{Value}->{$1} = $2;
      $self->{FileData}->[$self->{LineCount}] = $_;
    }
    elsif(s{^\s*(?!$self->{C})(\S+)\s*$self->{D}\s*(\S+)\s*$}
           {$1$self->{D}$2\n}) {
      $self->{Status}->{$1} = 1; # enabled
      $self->{Line}->{$1} = $self->{LineCount};
      $self->{Value}->{$1} = $2;
      $self->{FileData}->[$self->{LineCount}] = $_;
    }
    $self->{LineCount}++;
  }

}

# Update a key/value pair if it exists, and enables it in memory.  If the 
# key does not  If the key does not exists, it will be created if 'force' 
# is sent as the 3rd argument.  Otherwise, the error is fatal.

sub SetKey {

  my $self = shift;
  my $key = shift;
  my $value = shift;
  my $force = shift;

  if(exists $self->{Status}->{$key}) {
    $self->{Value}->{$key} = $value;
    $self->{FileData}->[$self->{Line}->{$key}] = "$key$self->{D}$value\n";
  }
  else {
    if($force && $force =~ /force/i) {
      $self->{Status}->{$key} = 1; # enabled
      $self->{Line}->{$key} = $self->{LineCount};
      $self->{LineCount}++;
      $self->{Value}->{$key} = $value;
      push @{$self->{FileData}}, "$key$self->{D}$value\n";
    }
    else {
      die("Key '$key' does not exist, force not set!");
    }
  }

}

# Updates the status of a key/value pair in memory.  A valid key must be
# specified and a valid status (0 - disabled, 1 - enabled), Otherwise the
# result will be fatal. 

sub SetKeyStatus {

  my $self = shift;
  my $key = shift;
  my $status = shift;

  # ensure the key exists.
  if(!defined($self->GetKey($key))) { die("'$key' is not a valid key."); }

  if($status == 1) {
    $self->{Status}->{$key} = $status;
    $self->{FileData}->[$self->{Line}->{$key}] =~
      s/^$self->{C}?(\S+)$self->{D}(\S+)\n$/$1$self->{D}$2\n/;
  }
  elsif($status == 0) {
    $self->{Status}->{$key} = $status;
    $self->{FileData}->[$self->{Line}->{$key}] =~
      s/^$self->{C}?(\S+)$self->{D}(\S+)\n$/$self->{C}$1$self->{D}$2\n/;
  }
  else {
    die("Invalid status argument given to SetKeyStatus.");
  }

  $self; 

}

# Determines the current value in memory of the given key.  If the key does
# not exist, undef is returned.

sub GetKey {

  my $self = shift;
  my $key = shift;

  return $self->{Value}->{$key};

}

# Determines the status (enabled, disabled) of a provided key.

sub GetKeyStatus {

  my $self = shift;
  my $key = shift;

  return $self->{Status}->{$key};

}

# PRIVATE:  This method synchronizes the changes made to a configration file
# in memory with the copy on disk.  Note:  This method is automatically
# called when the object is destroyed. 

sub _commit {

  my $self = shift;

  # Note on in-place configuration file updating:
  # Reading everything into memory works for small files, but it
  # doesn't scale well.  Trying it on your 2 GB web server log files
  # will wreak havoc on your machines VM system.  For small
  # configuration files though, this will be fine.

  # Although this algorithm is O(n^2) if the number of parameters
  # being changed is large, the advantage is that the changes to
  # the configuration file are very un-intrustive, meaning if the
  # user adds their own changes to the file somewhere, those
  # changes will not be clobbered by this script.  It is very
  # difficult to be un-intrusive if the configuration file is
  # regenerated from scratch every time.

  my $fh = $self->{FH};

  seek($self->{FH}, 0, 0) or die "can't seek to start of $self->{Filename}: $!";
  print $fh @{$self->{FileData}} or die "can't print to $self->{Filename}: $!";
  truncate($self->{FH}, tell($self->{FH})) 
    or die "can't truncate $self->{Filename}: $!";
  close($self->{FH}) or die "can't close $self->{Filename}: $!";

}

# PRIVATE:  This default constructor will be called when self 
# falls out of scope and is destroyed by the perl interpreter.

sub DESTROY {

  my $self = shift;

  $self->_commit;

}

# DEBUG:  This method is dumps the contents of $self->{FileData} which is
# just the configuration file being manipulated in memory.

sub _dump {

  my $self = shift;

  print @{$self->{FileData}};

}

1;
